// Take an observable and N callbacks, and observe the observable,
// ensuring it is called exactly N times, resolving once it has done so.
// Optionally takes a timeout, which it will wait X ms after the Nth callback
// to ensure it is not called again.
export function observableToPromiseAndSubscription(_a) {
    var observable = _a.observable, _b = _a.shouldResolve, shouldResolve = _b === void 0 ? true : _b, _c = _a.wait, wait = _c === void 0 ? -1 : _c, _d = _a.errorCallbacks, errorCallbacks = _d === void 0 ? [] : _d;
    var cbs = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        cbs[_i - 1] = arguments[_i];
    }
    var subscription = null;
    var promise = new Promise(function (resolve, reject) {
        var errorIndex = 0;
        var cbIndex = 0;
        var results = [];
        var tryToResolve = function () {
            if (!shouldResolve) {
                return;
            }
            var done = function () {
                subscription.unsubscribe();
                // XXX: we could pass a few other things out here?
                resolve(results);
            };
            if (cbIndex === cbs.length && errorIndex === errorCallbacks.length) {
                if (wait === -1) {
                    done();
                }
                else {
                    setTimeout(done, wait);
                }
            }
        };
        subscription = observable.subscribe({
            next: function (result) {
                var cb = cbs[cbIndex++];
                if (cb) {
                    try {
                        results.push(cb(result));
                    }
                    catch (e) {
                        return reject(e);
                    }
                    tryToResolve();
                }
                else {
                    reject(new Error("Observable called more than " + cbs.length + " times"));
                }
            },
            error: function (error) {
                var errorCb = errorCallbacks[errorIndex++];
                if (errorCb) {
                    try {
                        // XXX: should we collect these results too?
                        errorCb(error);
                    }
                    catch (e) {
                        return reject(e);
                    }
                    tryToResolve();
                }
                else {
                    reject(error);
                }
            },
        });
    });
    return {
        promise: promise,
        subscription: subscription,
    };
}
export default function (options) {
    var cbs = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        cbs[_i - 1] = arguments[_i];
    }
    return observableToPromiseAndSubscription.apply(void 0, [options].concat(cbs)).promise;
}
//# sourceMappingURL=observableToPromise.js.map